/*
   Pseudo

   Luc Maranget, projet Moscova, Inria Rocquencourt
 
   Copyright 2006 Institut National de Recherche en Informatique et    
   en Automatique.  All rights reserved.  This file is distributed     
   under the terms of the Q Public License version 1.0.                
*/

class Hash {
  private AList [] t ;
  private int sz ;
  private int n ;

  private void rehash() {
    int sz2 = 2*sz ;
    AList [] t2 = new AList [sz2] ;
    for (int k = 0 ; k < sz ; k++) {
      for (AList p = t[k] ; p != null ; p = p.next) {
	int h = hash(p.key) % sz2 ;
	t2[h] = new AList (p.key, p.val, t2[h]) ;
      }
    }
    sz = sz2 ;
    t = t2 ;
  }

  Hash() {
    sz = 4 ;
    t = new AList [sz] ;
  }

  private static char hash(char key) {
    key += ~(key << 7);
    key ^= (key >>> 1);
    key += (key << 2);
    key ^=  (key >>> 7);
    key += ~(key << 4);
    key ^=  (key >>> 8);
    return key;
  }

  
  int get(char key) {
    int h = hash(key) % sz ;
    AList p = AList.lookCell(key, t[h]) ;
    if (p == null) {
      return -1 ;
    } else {
      return p.val ;
    }
  }

  void set(char key, char val) {
    char h = hash(key) ;
    int k = h % sz ;
    AList p = AList.lookCell(key, t[k]) ;
    if (p == null) {
      n++ ;
      while (n > sz) { rehash() ; }
      k = h % sz ;
      p = new AList(key, val, t[k]) ;
      t[k] = p ;
    }
  }

  char [] inverse() {
    char [] r = new char[n] ;
    for (int k = 0 ; k < sz ; k++) {
       for (AList p = t[k] ; p != null ; p = p.next) {
         r[p.val] = p.key ;
       }
    }
    return r ;
  }

  int card() { return n ; }

  public String toString() {
     StringBuffer b = new StringBuffer () ;
     for (int k = 0 ; k < sz ; k++) {
       for (AList p = t[k] ; p != null ; p = p.next) {
	 b.append(p.key + "[" + (int)p.val + "]") ;
       }
     }
     return b.toString() ;
  }
}
