/*
   Pseudo

   Luc Maranget, projet Moscova, Inria Rocquencourt
 
   Copyright 2006 Institut National de Recherche en Informatique et    
   en Automatique.  All rights reserved.  This file is distributed     
   under the terms of the Q Public License version 1.0.                
*/

import java.io.* ;
import java.util.* ;

class Tree {
  char val ;
  Tree down, right ;
  int count ;
  int eow ;

  Tree (char val) {
    this(val, null, null) ;
  }

  Tree(char val, Tree down, Tree right) {
    this.val = val ; this.down = down ; this.right = right ;
    count = eow = 0 ;
  }

  /* Insert new words in trees */
  private static Tree insertDown(Tree t, String w, int idx, boolean eow) {
    if (idx+1 >= w.length()) {
      if (eow) { t.eow++; }
      else{ t.count++ ; }
      return t.down ;
    } else {
      return insertAux(t.down, w.charAt(idx+1), w, idx+1, eow) ;
    }
  }
  private static Tree insertAux
    (Tree t, char x, String w, int idx, boolean eow) {
    if (t == null) {
      t = new Tree (x) ;
    }
    char y = t.val ;
    if (y < x) {
      t.right = insertAux(t.right, x, w, idx, eow) ;
      return t ;
    } else if (y > x) {
      t = new Tree (x, null, t) ;
      t.down = insertDown(t, w, idx, eow) ;
      return t ;
    } else { // x == y
      t.down = insertDown(t, w, idx, eow) ;
      return t ;
    }
  }

  static Tree insert(Tree t, String w, boolean eow) {
    return insertAux(t, w.charAt(0), w, 0, eow) ;
  }

  /* find subtree for given prefix */
  private static Tree find(Tree t, String w, int idx) {
    char x = w.charAt(idx) ;
    while (t != null) {
      if (x == t.val) {
        idx++ ;
        if (idx >= w.length()) {
          return t ;
        } else {
          x = w.charAt(idx) ;
          t = t.down ;
        }
      } else if (x < t.val) {
        return null ;
      } else {
        t = t.right ;
      }
    }
    return null ;
  }

  static Tree find(Tree t, String w) {
    return find(t, w, 0) ;
  }
 
  static boolean exists(Tree t, String w) {
    Tree found = find(t, w) ;
    return found != null && found.count > 0 ;
  }
  
  /* Put up links */

  private static void putUpNext(Tree t, Tree top, StringBuffer b) {
    for ( ; t != null ; t = t.right) {
      // Find branch in top tree
      while (top != null && top.val < t.val) { top = top.right ; }
      if (top == null) { return ; }
      if (top.val > t.val) { continue ; }

      b.append(t.val) ;
      putUpNext(t.down, top.down, b) ;
      if (t.count > 0) {
        if (t.down != null) {
          System.err.println
            ("UP AND DOWN: " + b + "[" + coder.decode(b.substring(1)) + "]") ;
          System.exit(2) ;
        }
        t.down = top.down ;
      }
      b.setLength(b.length()-1) ;
    }
  }

  private static void putUpFirst(Tree t, Tree top, StringBuffer b) {
    for ( ; t != null ; t = t.right) {
      b.append(t.val) ;
      putUpNext(t.down, top, b) ;
      b.setLength(b.length()-1) ;
    }
  }

  private static void putUp(Tree t, Tree top) {
    putUpFirst(t, top, new StringBuffer()) ;
  }

  /* Simplification */
  private static boolean changed ;

  private static Tree simplifTree(Tree t) {
    if (t == null) {
      return null ;
    } else {
      Tree right = simplifTree(t.right) ;
      if (t.count > 0 || t.eow > 0) {
        if (t.eow == 0 && t.down == null) {
          changed = true ;
          return right ;
        } else {
          t.right = right ;
          t.down = null ;
          return t ;
        }
      } else {
        t.down = simplifTree(t.down) ;
        if (t.down == null) {
          return right ;
        } else {
          t.right = right ;
          return t ;
        }
      }
    }
  }

  static int verbose ;

  static Tree simplif(Tree start, Tree tw) {
    int count = 1 ;
    StringBuffer b = new StringBuffer() ;
    do {
      changed = false ;
      if (verbose > 0) {
        System.err.println("COUNT: " + count++) ;
        if (verbose > 1) {
          System.err.println("START: ") ;
          dump(System.err, start) ;
          System.err.println("TW: ") ;
          dump(System.err, tw) ;
        }
      }
      putUp(start, tw) ; putUp(tw, tw) ;
      start = simplifTree(start) ;
      tw = simplifTree(tw) ;
    } while (changed) ;
    putUp(start, tw) ; putUp(tw, tw) ;
    return start ;
  }
  
  
  /* Dump trees */
  static Coder coder ;

  private static void dump(PrintWriter o, Tree t, StringBuffer b) {
    if (t == null) return ;
    for ( ; t != null ; t = t.right) {
      b.append(coder.decode(t.val)) ;
      if (t.count > 0 || t.eow > 0) {
        o.println(b + (t.eow > 0 ? "${" + t.eow + "}"  : "") +
                  "[" + t.count + "]") ;
      }
      dump(o, t.down, b) ;
      b.setLength(b.length()-1) ;
    }
  }


  static void dump(PrintWriter o, Tree t) {
    dump(o, t, new StringBuffer()) ;
  }

  static void dump(PrintStream o, Tree t) {
    PrintWriter oo = new PrintWriter(o) ;
    dump(oo, t) ;
    oo.flush() ;
  }

  static void dump(Tree t) {
    dump(System.out, t) ;
  }
}
